@@ -116,23 +116,35 @@ let enable_outside_detected_project =
       (List.map File_system.project_root_witness ~f:(fun name ->
            Format.sprintf "$(b,%s)" name ) )
   in
-  let doc =
-    Format.sprintf
-      "Read $(b,.ocamlformat) config files outside the current project when \
-       no project root has been detected for the input file. The project \
-       root of an input file is taken to be the nearest ancestor directory \
-       that contains a %s file. If $(b,.ocamlformat) config files are \
-       located in the same directory or parents they are applied, if no \
-       $(b,.ocamlformat) file is found then the global configuration \
-       defined in $(b,\\$XDG_CONFIG_HOME/.ocamlformat) (or in \
-       $(b,\\$HOME/.config/.ocamlformat) if $(b,\\$XDG_CONFIG_HOME) is \
-       undefined) is applied."
-      witness
+  let enable =
+    let doc_enable =
+      Format.sprintf
+        "Read $(b,.ocamlformat) config files outside the current project \
+         when no project root has been detected for the input file. The \
+         project root of an input file is taken to be the nearest ancestor \
+         directory that contains a %s file. If $(b,.ocamlformat) config \
+         files are located in the same directory or parents they are \
+         applied, if no $(b,.ocamlformat) file is found then the global \
+         configuration defined in $(b,\\$XDG_CONFIG_HOME/.ocamlformat) (or \
+         in $(b,\\$HOME/.config/.ocamlformat) if $(b,\\$XDG_CONFIG_HOME) is \
+         undefined) is applied."
+        witness
+    in
+    Arg.info ["enable-outside-detected-project"] ~doc:doc_enable ~docs
+  in
+  let disable =
+    let doc_disable =
+      "If no $(b,.ocamlformat) config files have been detected, disable the \
+       formatting. OCamlFormat is disabled outside of a detected project by \
+       default, to enable the opposite behavior use \
+       $(b,--enable-outside-detected-project)."
+    in
+    Arg.info ["disable-outside-detected-project"] ~doc:doc_disable ~docs
   in
   Term.(
     const (fun enable_outside_detected_project conf ->
         {conf with enable_outside_detected_project} )
-    $ Arg.(value & flag & info ["enable-outside-detected-project"] ~doc ~docs) )
+    $ Arg.(value & vflag false [(true, enable); (false, disable)]) )
 
 let inplace =
   let doc = "Format in-place, overwriting input file(s)." in
@@ -615,6 +627,10 @@ type action =
   | Check of input list
   | Print_config of Conf.t
 
+let ( let* ) = Result.( >>= )
+
+let ( let+ ) = Result.( >>| )
+
 let make_action ~enable_outside_detected_project ~root action inputs =
   let make_file ?name kind file =
     let name = Option.value ~default:file name in
@@ -622,14 +638,14 @@ let make_action ~enable_outside_detected_project ~root action inputs =
       build_config ~enable_outside_detected_project ~root ~file:name
         ~is_stdin:false
     in
-    Ok {kind; name; file= File file; conf}
+    {kind; name; file= File file; conf}
   in
   let make_stdin ?(name = "<standard input>") kind =
     let+ conf =
       build_config ~enable_outside_detected_project ~root ~file:name
         ~is_stdin:false
     in
-    Ok {kind; name; file= Stdin; conf}
+    {kind; name; file= Stdin; conf}
   in
   let make_input = function
     | `Single_file (kind, name, f) -> make_file ?name kind f
@@ -638,15 +654,15 @@ let make_action ~enable_outside_detected_project ~root action inputs =
   let make_inputs = function
     | (`Single_file _ | `Stdin _) as inp ->
         let+ inp = make_input inp in
-        Ok [inp]
+        [inp]
     | `Several_files files ->
         let+ inputs =
           List.fold_left files ~init:(Ok []) ~f:(fun acc (kind, file) ->
-              let+ acc = acc in
+              let* acc = acc in
               let+ file = make_file kind file in
-              Ok (file :: acc) )
+              file :: acc )
         in
-        Ok (List.rev inputs)
+        List.rev inputs
   in
   match (action, inputs) with
   | `Print_config, inputs ->
@@ -661,7 +677,7 @@ let make_action ~enable_outside_detected_project ~root action inputs =
       let+ conf =
         build_config ~enable_outside_detected_project ~root ~file ~is_stdin
       in
-      Ok (Print_config conf)
+      Print_config conf
   | (`No_action | `Output _ | `Inplace | `Check), `No_input ->
       Error "Must specify at least one input file, or `-` for stdin"
   | (`No_action | `Output _), `Several_files _ ->
@@ -671,16 +687,16 @@ let make_action ~enable_outside_detected_project ~root action inputs =
       Error "Cannot specify stdin together with --inplace"
   | `No_action, ((`Single_file _ | `Stdin _) as inp) ->
       let+ inp = make_input inp in
-      Ok (In_out (inp, None))
+      In_out (inp, None)
   | `Output output, ((`Single_file _ | `Stdin _) as inp) ->
       let+ inp = make_input inp in
-      Ok (In_out (inp, Some output))
+      In_out (inp, Some output)
   | `Inplace, ((`Single_file _ | `Several_files _) as inputs) ->
       let+ inputs = make_inputs inputs in
-      Ok (Inplace inputs)
+      Inplace inputs
   | `Check, ((`Single_file _ | `Several_files _ | `Stdin _) as inputs) ->
       let+ inputs = make_inputs inputs in
-      Ok (Check inputs)
+      Check inputs
 
 let validate_inputs () =
   match (!global_conf.inputs, !global_conf.kind, !global_conf.name) with
@@ -742,8 +758,8 @@ let validate () =
     !global_conf.enable_outside_detected_project && Option.is_none root
   in
   match
-    let+ action = validate_action () in
-    let+ inputs = validate_inputs () in
+    let* action = validate_action () in
+    let* inputs = validate_inputs () in
     make_action ~enable_outside_detected_project ~root action inputs
   with
   | Error e -> `Error (false, e)
