@@ -1,6 +1,6 @@
 let foo = {%foo| foooooooooooooo |}
 
-let foo = (* A *) ({%foo| foooooooooooooo |} (* B *) [@attr]) (* C *)
+let foo = (* A *) {%foo| foooooooooooooo |} (* B *) [@attr] (* C *)
 
 let foo = (* A *) {%foo sep| foooooooooooooo |sep}
 
@@ -8,11 +8,11 @@ let foo = {%foo| foooooooooooooo |} [@@attr]
 
 let foo = {%foo| foooooooooooooo |} (* A *) [@@attr] (* B *)
 
-let foo = ({%foo| foooooooooooooo |} [@attr]) [@@attr]
+let foo = {%foo| foooooooooooooo |} [@attr] [@@attr]
 
-let foo = ({%foo| foooooooooooooo |} (* A *) [@attr]) (* B *) [@@attr]
+let foo = {%foo| foooooooooooooo |} (* A *) [@attr] (* B *) [@@attr]
 
-let foo = (* A *) ({%foo| foooooooooooooo |} [@attr]) (* B *) [@@attr]
+let foo = (* A *) {%foo| foooooooooooooo |} [@attr] (* B *) [@@attr]
 
 let foo = (* A *) {%foo sep| foooooooooooooo |sep} (* B *) [@@attr]
 
